@@ -1,10 +1,9 @@
 from contextlib import suppress
 
-import pandas as pd
 import psycopg2 as pgsql
 from pydantic import Field, SecretStr, constr, create_model
 
-from toucan_connectors.common import adapt_param_type, convert_to_printf_templating_style
+from toucan_connectors.common import pandas_read_sql
 from toucan_connectors.toucan_connector import ToucanConnector, ToucanDataSource, strlist_to_enum
 
 
@@ -113,8 +112,9 @@ class PostgresConnector(ToucanConnector):
         connection = pgsql.connect(**self.get_connection_params(database=data_source.database))
 
         query_params = data_source.parameters or {}
-        query = convert_to_printf_templating_style(data_source.query)
-        df = pd.read_sql(query, con=connection, params=adapt_param_type(query_params))
+        df = pandas_read_sql(
+            data_source.query, con=connection, params=query_params, adapt_params=True
+        )
 
         connection.close()
 
